/* outletsrv - see szInfo below for copyright and licence details */

const char szInfo[] = "\
Remote Outlet Control Server\n\
Version 1.0\n\
\n\
A server to control remote outlets via USB serial.\n\
See http://www.sconemad.com for more information.\n\
\n\
Copyright (c) 2009 Andrew Wedgbury <wedge@sconemad.com>\n\
\n\
This program is free software; you can redistribute it and/or modify\n\
it under the terms of the GNU General Public License as published by\n\
the Free Software Foundation; either version 2, or (at your option)\n\
any later version.\n\
\n\
This program is distributed in the hope that it will be useful,\n\
but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
GNU General Public License for more details.\n\
\n\
You should have received a copy of the GNU General Public License\n\
along with this program (see the file COPYING); if not, write to the\n\
Free Software Foundation, Inc.,\n\
59 Temple Place - Suite 330, Boston, MA  02111-1307, USA\n\
";

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <string.h>
#include <string>
#include <iostream>
#include <getopt.h>
#include <errno.h>

// Command line option defaults:

static bool opt_debug = false;

// Default network interface on which to listen for connections
//
// This can be either the IP address of a specific interface,
// or:
// "0.0.0.0" to allow connections from anywhere
// "127.0.0.1" to allow connections from only the local machine
//
std::string opt_addr = "0.0.0.0";

// Default (UDP) port on which to listen
//
// I picked 7900 at random as the default.
// Note that if you change this, you also need to change in in the client.
//
short opt_port = 7900;

// Com port
//
// This default is used if one is not specified on the command line
//
std::string comport = "/dev/ttyACM0";

int g_sock;
int g_hcom;

//=============================================================================
void exit_error(const std::string& msg)
{
  std::cerr << "outletsrv: " << msg
            << " (errno=" << errno << ")"
            << "\n";
  exit(1);
}

//=============================================================================
int open_socket()
{
  g_sock = socket(PF_INET,SOCK_DGRAM,IPPROTO_UDP);
  if (g_sock < 0) {
    exit_error("socket() failed");
  }

  sockaddr_in sa;
  memset(&sa,0,sizeof(sa));
  sa.sin_family = AF_INET;
  if (inet_pton(AF_INET,opt_addr.c_str(),&sa.sin_addr.s_addr) <= 0) {
    exit_error("inet_pton() failed - check address is valid");
  }
  sa.sin_port = htons(opt_port);
   
  if (bind(g_sock,(const struct sockaddr*)&sa,sizeof(sa)) < 0) {
    exit_error("bind() failed - check address and port is valid and you have the correct permissions");
  }

  if (opt_debug) {
    std::cout << "outletsrv waiting for data on "
              << opt_addr << ":"
              << opt_port
              << "\n";
  }

  return 0;
}

//=============================================================================
int open_comms(const char* port)
{
  g_hcom = open(port,O_RDWR,0);
  if (g_hcom < 0) {
    exit_error("open comm port failed");
  }
}

//=============================================================================
int run()
{
  while (true) {
    char buffer[64];
    
    sockaddr_in client_sa;
    socklen_t client_sa_size = sizeof(client_sa);
    memset(&client_sa,0,client_sa_size);

    int len = recvfrom(g_sock,&buffer,sizeof(buffer),0,(struct sockaddr*)&client_sa,&client_sa_size);
    
    if (len <= 0) {
      if (opt_debug) {
        std::cout << "recvfrom() failed\n";
      }
      continue;
    }

    buffer[len] = '\0';
    
    char str[INET_ADDRSTRLEN];
    const char* client_addr = inet_ntop(AF_INET, &client_sa.sin_addr, str, sizeof(str));
    if (opt_debug) {
      std::cout << "Received data from " << client_addr << ": " << buffer << "\n";
    }

    if (len != write(g_hcom,buffer,len)) {
      std::cout << "Write to comm port failed\n";
    }
    
  }

  return 0;
}


//=============================================================================
int main(int argc,char* argv[])
{
  // Look for version/help options initially as a special case
  if (argc == 2) {
    std::string arg = argv[1];
    bool opt_version = (arg == "-v" || arg == "--version");
    bool opt_help = (arg == "-h" || arg == "--help");

    if (opt_version) {
      std::cout << szInfo;
      return 0;
    }

    if (opt_help) {
      std::cout
        << "Usage: outletsrv [OPTIONS]\n"
        << "\n"
        << "Command line options:\n"
        << " -h         Print this help and exit\n"
        << " -v         Print version information and exit\n"
        << " -d         Debug mode (run in foreground and print status to console)\n"
        << " -c=COMPORT Specify comm port (default: /dev/ttyACM0)\n" 
        << "\n";
      return 0;
    }
  }

  int c;
  extern char* optarg;
  while ((c = getopt(argc,argv,"dc:")) >= 0) {
    switch (c) {
      case '?': return 1;
      case 'd': opt_debug = true;  break;
      case 'c': comport = optarg;  break;
    }
  }

  if (!opt_debug) {
    // Daemonize if not in debug mode
    pid_t pid;
    if ((pid=fork()) != 0) {
      exit(0);
    }
    setsid();
    if ((pid=fork()) != 0) {
      exit(0);
    }
    umask(0);
  }

  open_socket();
  open_comms(comport.c_str());
  
  run();

  close(g_sock);
  close(g_hcom);

  return 0;
}
